"""\
Pandas Selector
---------------

Simple, composable column selector for ``loc[]``, ``iloc[]``, ``assign()`` and others. Juse use ``DF``.

Filter rows:

>>> from pandas_selector import DF
>>> df = pd.DataFrame({"x": range(9), "y": 3 * ["a", "B", "c"]})
>>> df.loc[DF.x < 3]
   x  y
0  0  a
1  1  B
2  2  c

Access nested column attributes:

>>> df.loc[DF.y.str.islower()]
   x  y
0  0  a
2  2  c
3  3  a
5  5  c
6  6  a
8  8  c

Combine filter predicates:

>>> df.loc[DF.y.str.islower() & (df.x < 3)]
   x  y
0  0  a
2  2  c

Create new columns:

>>> df.assign(z = DF.x * DF.y)
   x  y         z
0  0  a
1  1  B         B
2  2  c        cc
3  3  a       aaa
4  4  B      BBBB
5  5  c     ccccc
6  6  a    aaaaaa
7  7  B   BBBBBBB
8  8  c  cccccccc

Chain operations:

>>> (df
...  .assign(z = DF.x * DF.y)
...  .loc[DF.z.str.len() > 3]
... )
   x  y         z
3  3  a       aaa
4  4  B      BBBB
5  5  c     ccccc
6  6  a    aaaaaa
7  7  B   BBBBBBB
8  8  c  cccccccc


Author: Eike von Seggern <eike@vonseggern.space>
"""
__version__ = "0.1.1"
__all__ = ["DF"]

from .accessor import DataframeAccessor

DF = DataframeAccessor()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pandas_selector']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1,<2']

extras_require = \
{'rtdbuild': ['sphinx>=3.5.0,<4.0.0', 'pydata-sphinx-theme>=0.6.0,<0.7.0']}

setup_kwargs = {
    'name': 'pandas-selector',
    'version': '0.1.1',
    'description': 'Simple, composable column selector for loc[], iloc[], assign() and others.',
    'long_description': 'Pandas Selector\n===============\n\nSimple, composable column selector for ``loc[]``, ``iloc[]``, ``assign()`` and others.\n\n.. image:: https://readthedocs.org/projects/pandas-selector/badge/?version=latest\n  :target: https://pandas-selector.readthedocs.io/en/latest/?badge=latest\n  :alt: Documentation Status\n.. image:: https://github.com/eikevons/pandas-selector/actions/workflows/check.yml/badge.svg\n  :target: https://github.com/eikevons/pandas-selector/actions/workflows/check.yml\n  :alt: Test Status\n\nOverview\n--------\n\n- **Motivation**: Make chaining Pandas operations easier and bring\n  functionality to Pandas similar to Spark\'s `col()\n  <https://spark.apache.org/docs/latest/api/python/reference/api/pyspark.sql.functions.col.html#pyspark.sql.functions.col>`_\n  function or referencing columns in R\'s `dplyr\n  <https://dplyr.tidyverse.org/articles/dplyr.html>`_.\n- **Install** from PyPI with ``pip install\n  pandas-selector``. Pandas versions 1.0+ (``^1.0``) are supported.\n- **Documentation** can be found at `readthedocs\n  <https://readthedocs.org/projects/pandas-selector/>`_.\n- **Source code** can be obtained from `GitHub <https://github.com/eikevons/pandas-selector>`_.\n\nExample: Create new column and filter\n-------------------------------------\n\nInstead of writing "traditional" Pandas like this:\n\n.. code-block:: python\n\n    df_in = pd.DataFrame({"x": range(5)})\n    df = df_in.assign(y = df_in.x // 2)\n    df = df.loc[df.y <= 1]\n    df\n    #    x  y\n    # 0  0  0\n    # 1  1  0\n    # 2  2  1\n    # 3  3  1\n\nOne can write:\n\n.. code-block:: python\n\n    from pandas_selector import DF\n    df = (df_in\n          .assign(y = DF.x // 2)\n          .loc[DF.y <= 1]\n         )\n\nThis is especially handy when re-iterating on data frame manipulations\ninteractively, e.g. in a notebook.\n\nBut you can access all methods and attributes of the data frame from the\ncontext:\n\n.. code-block:: python\n\n    df = pd.DataFrame({\n        "X": range(5),\n        "Y": ["1", "a", "c", "D", "e"],\n    })\n    df.loc[DF.y.str.isupper() | DF.y.str.isnumeric()]\n    #    X  y\n    # 0  0  1\n    # 3  3  D\n    df.loc[:, DF.columns.str.isupper()]\n    #    X\n    # 0  0\n    # 1  1\n    # 2  2\n    # 3  3\n    # 4  4\n\nSimilar projects for pandas\n===========================\n\n* `pandas-ply <https://github.com/coursera/pandas-ply>`_\n\n  * (-) stale(?), last change 6 years ago\n  * (-) new API to learn\n  * (-) ``Symbol`` / ``pandas_ply.X`` works only with ``ply_*`` functions\n\n* `pandas-select <https://pandas-select.readthedocs.io/en/latest/reference/label_selection.html>`_\n\n  * (+) no explicite ``df`` necessary\n  * (-) new API to learn\n\n* `pandas-selectable <https://github.com/jseabold/pandas-selectable>`_\n\n  * (+) simple ``select`` accessor\n  * (-) usage inside chains clumsy (needs explicite ``df``):\n\n    .. code-block:: python\n\n       ((df\n         .select.A == \'a\')\n         .select.B == \'b\'\n       )\n\n  * (-) hard-coded ``str``, ``dt`` accessor methods\n  * (?) composable?\n\nDevelopment\n===========\n\nDevelopment is containerized with [Docker](https://www.docker.com/) to\nseparte from host systems and improve reproducability. No other\nprerequisites are needed on the host system.\n\n**Recommendation for Windows users:** install `WSL 2\n<https://docs.microsoft.com/en-us/windows/wsl/install-win10>`_ (tested\non Ubuntu 20.04), and for containerized workflows, `Docker\nDesktop <https://www.docker.com/products/docker-desktop>`_ for Windows.\n\nThe **common tasks** are collected in ``Makefile`` (See ``make help`` for a\ncomplete list):\n\n- Run the unit tests: ``make test`` or ``make watch`` for continuously running\n  tests on code-changes.\n- Build the documentation: ``make docs``\n- **TODO**: Update the ``poetry.lock`` file: ``make lock``\n- Add a dependency:\n\n  1. Start a shell in a new container.\n  2. Add dependency with ``poetry add`` in the running container. This will update\n     ``poetry.lock`` automatically::\n\n        # 1. On the host system\n        % make shell\n        # 2. In the container instance:\n        I have no name!@7d0e85b3a303:/app$ poetry add --dev --lock falcon\n\n- Build the development image ``make devimage``\n  (Note: This should be done automatically for the targets.) \n',
    'author': 'Eike von Seggern',
    'author_email': 'eikevons@mailbox.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/eikevons/pandas-selector.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""Access wrappers for data frames members."""
from typing import Callable, Optional, Sequence, Union

import pandas as pd

# Wrappers for attribute, item and operator access
class WrapperBase:
    """Base class for wrapping attribute, item or operator access."""
    def __init__(self, name: str):
        """
        Parameters
        ----------
        name
            Name of the accessed object member.
        """
        self.name = name

    def __repr__(self):
        return f"<{type(self).__name__} {self.name}>"

    def __call__(self, obj, root_df: pd.DataFrame):
        """Access member of wrapped object.

        Parameters
        ----------
        obj
            Typically the ``~pandas.DataFrame`` or ``~pandas.Series`` to act
            upon.
        root_df
            The original data frame from the context.
        """
        raise NotImplementedError("Must be implemented by a sub-class.")

class Attribute(WrapperBase):
    """Wrap ``df.column_name`` or similar access patterns."""
    def __call__(self, obj, root_df):
        """Access attribute of ``obj``.

        Parameters
        ----------
        obj
            Typically the ``~pandas.DataFrame`` or ``~pandas.Series`` to act
            upon.
        root_df
            The original data frame from the context. (Unused)
        """
        return getattr(obj, self.name)

    def __str__(self):
        return f".{self.name}"


class Item(WrapperBase):
    """Wrap ``df["column_name"]`` or similar access patterns."""
    def __call__(self, obj, root_df):
        return obj[self.name]

    def __str__(self):
        return f"[{self.name!r}]"

class Method(WrapperBase):
    """Wrap method and operator calls.

    E.g.

    * ``DF.x <= other`` or similar operator patterns
    * ``DF.x.method(...)``
    * :code:`DF.x.method(1 + 'abc')`
    """
    def __init__(self, name: str,
                 # this: "DataframeAccessor",
                 *args, **kwargs):
        """
        Parameters
        ----------
        name
            Operator name, e.g. `__eq__`.
        this
            ``DataframeAccessor`` to apply operator upon.
        *args, **kwargs
            Optional arguments for the operator, e.g. the second argument for a binary operator.
        """
        super().__init__(name)
        # self.this = this
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        # TODO include args[0] (other) in string where sensible
        return f"<{type(self).__name__}: {self.name}(...)>"

    # TODO include args[0] (other) in string where sensible
    def __str__(self):
        # return f"{self.this}.{self.name}(...)"
        return f".{self.name}(...)"

    def __call__(self, obj, root_df):
        # this = self.this(obj)
        op_meth = getattr(obj, self.name)
        if len(self.args) == 1 and isinstance(self.args[0], DataframeAccessor):
            other = self.args[0](root_df)
            return op_meth(other)
        return op_meth(*self.args, **self.kwargs)


def _add_dunder_operators(cls):
    """Dress class with all sensible comparison operations.

    The class must implement a ``_operator_proxy`` method.
    """
    for op in [
        "__abs__",
        "__add__",
        "__and__",
        "__bool__",
        "__contains__",
        "__div__",
        "__divmod__",
        "__eq__",
        "__floordiv__",
        "__ge__",
        "__gt__",
        "__invert__",
        "__le__",
        "__lt__",
        "__mul__",
        "__ne__",
        "__neg__",
        "__not__",
        "__or__",
        "__pos__",
        "__pow__",
        "__sub__",
        "__truediv__",
        "__xor__",
        ]:
        # Fix the closure of `op_wrap` to the current value of `op`. Without
        # `fix_closure()` all created methods point to the last `op` value.
        def fix_closure(op):
            def op_wrap(self, *args, **kwargs):
                return self._operator_proxy(op)(*args, **kwargs)
            return op_wrap
        setattr(cls, op, fix_closure(op))
    return cls


@_add_dunder_operators # This is necessary to overload all dunder operators.
class DataframeAccessor:
    """Build callable for column access and operators.

    Use the global instance like::

        from pandas_selector import DF
        df.loc[DF.x < 3]

    All operations (item/attribute access, method calls) are passed to the
    data frame of the context.

    This is useful in combination with :attr:`~pandas.DataFrame.loc`,
    :attr:`~pandas.DataFrame.iloc`, :meth:`~pandas.DataFrame.assign` and
    other methods that accept callables taking the data frame to act on as
    single argument.

    Examples
    --------
    Usage with ``loc`` or ``iloc``:

    >>> DF = DataframeAccessor()
    >>> df = pd.DataFrame({"x": [1, 2, 3, 4]})
    >>> df.loc[DF.x <= 2]

    Usage with ``assign()``:

    >>> df.assign(y = DF.x * 2)
    >>> df
    """
    def __init__(self,
                 levels:Union[Sequence[Callable], None]=None):
        """
        Parameters
        ----------
        levels:
            Sequence of callables to extract attributes from data frames or series.
        """
        self._levels = levels or ()

    def __repr__(self):
        return f"<{type(self).__name__} {'.'.join(repr(l) for l in self._levels)}>"

    def __str__(self):
        return ''.join(str(l) for l in self._levels)

    def __getattr__(self, name):
        return DataframeAccessor(self._levels + (Attribute(name),))

    def __getitem__(self, key):
        return DataframeAccessor(self._levels + (Item(key),))

    def _operator_proxy(self, op_name):
        """Generate proxy function for built-in operators.

        Used by :func:`_add_dunder_operators`
        """
        def op_wrapper(*args, **kwargs):
            return DataframeAccessor(self._levels + (Method(op_name, *args, **kwargs),))
        return op_wrapper

    def __call__(self, *args, **kwargs):
        # Heuristic: Assume the selector is applied if exactly one DataFrame
        # argument is passed.
        if len(args) == 1 and isinstance(args[0], pd.DataFrame):
            obj = root_df = args[0]
            for lvl in self._levels:
                obj = lvl(obj, root_df)
            return obj

        # Create a new DataframeAccessor
        return DataframeAccessor(self._levels[:-1] + (Method(self._levels[-1].name, *args, **kwargs),))
